'' =================================================================================================
''
''   File....... jm_parser.spin
''   Purpose.... 
''   Author..... Jon "JonnyMac" McPhalen
''               Copyright (c) 2015 Jon McPhalen
''               -- see below for terms of use
''   E-mail..... jon@jonmcphalen.com
''   Started.... 
''   Updated.... 23 APR 2015
''
'' =================================================================================================

' 23 APR 2015 -- fixed enqueue_str for strings w/o CR or LF
' 25 MAR 2015 -- changed ucase to accept token index instead of string pointer
' 12 MAR 2015 -- updated tokens to 12  
' 11 MAR 2015 -- changed token length to accommdate indicated binary up to 32 bits


con

  BUF_SIZE = 80 + 1                                              ' size of input buffer + 0 terminator
  TOK_SIZE = 34 + 1                                              ' size of each token + 0 terminator

  N_TOKENS = 12                                                  ' allowable tokens in string


obj

  str : "jm_strings"


var

  byte  queue[BUF_SIZE]                                          ' space for incoming characters
  byte  qidx                                                     ' buffer index
  byte  tokens[TOK_SIZE * N_TOKENS]                              ' space for parsed tokens

  long  p_vchars                                                 ' pointer to valid characters
  long  p_vtlist                                                 ' pointer to valid tokens list
  long  ntokens                                                  ' number of tokens in list
  long  csensitive

  long  tcount                                                   ' tokens in string


pub start(p_valid, p_tokens, ntoks, cs)

'' Configure parser
'' -- p_valid is pointer to string of valid characters
'' -- p_token is pointer to first token in valid token list
'' -- ntoks in number of tokesn in list
'' -- cs is case sensitivity (when false everything converted to uppercase)

  longmove(@p_vchars, @p_valid, 4)                               ' copy parameters

  reset                                                          ' clear everything


pub reset
                        
  bytefill(@queue, 0, BUF_SIZE + 1)                              ' clear string spaces
  bytefill(@tokens, 0, TOK_SIZE * N_TOKENS)


pub enqueue(c)

'' Adds character c to input queue
'' -- supports backspace for live input

  case c
    0, 13, 10:                                                   ' end of input?
      queue[qidx] := 0                                           ' terminate input
      ifnot (csensitive)                                         ' not case sensitive?
        str.ucstr(@queue)                                        ' covert to uppercase
      tcount := count_tokens(@queue)                             ' possible tokens in queue?
      if (tcount > 0)
        extract_all
        return true
      else
        reset

    8:                                                           ' backspace?
      if (qidx > 0)                                              ' if chars in queue
        queue[--qidx] := 0                                       ' back up and erase last

    other:
      queue[qidx++] := c                                         ' add to buffer
      if (qidx == BUF_SIZE)
        qidx -= 1

  return false                                                   ' no tokens yet


pub enqueue_str(p_str)

'' Adds string to input queue
'' -- preformatted string should include terminator (CR)

  repeat strsize(p_str) + 1                                      ' include trailing 0 (if no CR or LF)
    if (enqueue(byte[p_str++]))
      return true

  return false
  

pub queue_addr

'' Returns hub address of queue for external .str() methods

  return @queue
  
  
pub queue_len

'' Returns current length of input queue

  return strsize(@queue)


pub count_tokens(p_str) | toks, tflag, c

'' Returns count of possible tokens in string
'' -- does not compare to apps's token list

  toks := 0                                                      ' zero tokens
  tflag := false                                                 ' not inside token

  repeat strsize(p_str)                                          ' iterate through string
    c := byte[p_str++]                                           ' get a character  
    if (str.first(c, p_vchars) => 0)                             ' c found in valid chars                                  
      if (tflag == false)                                        ' and not already in token
        tflag := true                                            ' mark in token
        ++toks                                                   ' new token, increment count
    else
      tflag := false                                             ' mark not in token
    
  return toks <# N_TOKENS


pub token_count

'' Returns token count
'' -- possible token count after count_tokens()
'' -- actual token count after extract_all 

  return tcount


pub extract_all | tidx, p_tok

'' Extract all (up to MAX_TOKENS) tokens from input queue

  bytefill(@tokens, 0, TOK_SIZE * N_TOKENS)                      ' clear tokens storage space

  tcount := count_tokens(@queue)                                 ' count tokens
  if (tcount == 0)                                               ' abort if no tokens in queue
    return

  repeat tidx from 0 to tcount-1
    extract_token(@queue, tidx)                                  ' extract token

  
pub extract_token(p_str, tidx) | target, tflag, c, p_tok, tlen

'' Extract token from buffer at p_str
'' -- tidx is the target token (0..N_TOKENS-1)

  if (tidx => N_TOKENS)                                          ' index is out-of-range
    return
  else
    target := tidx                                               ' copy token index

  tflag := false                                                 ' not inside token

  repeat strsize(p_str)                                          ' iterate through buffer
    c := byte[p_str++]                                           ' get a character
    if (str.first(c, p_vchars) < 0)                              ' c not valid?
      tflag := false                                             ' mark not in token
    else 
      if (tflag == false)                                        ' if not in token and letter or number
        tflag := true                                            ' mark in token
        if (target > 0)                                          ' if not at target
          --target                                               ' decrement 
        else                                                     ' token found - extract it
          p_tok := token_addr(tidx)                              ' point to token buffer     
          bytefill(p_tok, 0, TOK_SIZE)                           ' clear token buffer
          byte[p_tok++] := c                                     ' put current character into token buffer
          tlen := 1                                              ' length of current token
          repeat                                                 ' extract rest of token
            c := byte[p_str++]                                   ' get next character
            if (str.first(c, p_vchars) => 0)                     ' valid character
              if (++tlen < TOK_SIZE)                             ' room in token?
                byte[p_tok++] := c                               ' add to token buffer
              else
                return
            else
              return                                             ' we're done        


pub token_addr(tidx)

'' Returns start address of token string
'' -- can return address of unused token

  if ((tidx => 0) and (tidx < N_TOKENS))                         ' if in token storage space
    return @tokens + (tidx * TOK_SIZE)                           ' return hub adddress
  else
    return -1 


pub token_len(tidx)

'' Returns length of token at tidx

  return strsize(token_addr(tidx))
  
      
pub get_token_id(p_str) | p_tokens, tidx

'' Search known tokens list for string
'' -- returns index (0..ntokens-1) if found
'' -- returns -1 if no match

  p_tokens := p_vtlist                                           ' point to start of tokens list

  repeat tidx from 0 to ntokens-1                                ' iterate through known tokens
    if (strcomp(p_str, p_tokens))                                ' if match found
      return tidx                                                ' return token index
    else
      p_tokens += strsize(p_tokens) + 1                          ' advance to next token

  return -1


pub token_is_number(tidx)

'' Returns true if token at tidx is known number format
'' -- binary (%) and hex ($) values must be indicated (e.g., %1010 or $0A)

  return str.is_number(token_addr(tidx))    


pub token_value(tidx)

'' Returns value of token at tidx
'' -- 0 if not a number
'' -- use token_is_number() before if 0 is valid

  if (token_is_number(tidx))
    return str.asc2val(token_addr(tidx))
  else
    return 0    
     

pub value(p_str)

'' Convert string to value
'' -- must be correctly formatted

  return str.asc2val(p_str)


pub ucase(tidx)

'' Converts token at tidx to upper case

  str.ucstr(token_addr(tidx))


dat { license }

{{

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

}}  